import os
from StringIO import StringIO
import base64
import zlib
import uuid
import urllib
from sh import xmlsec1
from datetime import datetime
from lxml import etree
from lxml.builder import ElementMaker

XML_SIGNATURE_FRAGMENT = """
<ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#">
  <ds:SignedInfo>
    <ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>
    <ds:SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"/>
    <ds:Reference>
      <ds:Transforms>
        <ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>
        <ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>
      </ds:Transforms>
      <ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>
      <ds:DigestValue />
    </ds:Reference>
  </ds:SignedInfo>
  <ds:SignatureValue />
  <ds:KeyInfo>
    <ds:X509Data />
  </ds:KeyInfo>
</ds:Signature>
"""

def iso_no_microseconds(dt):
  return(datetime.isoformat(dt.replace(microsecond = 0)))

def gzip_and_base64encode(data):
  return base64.b64encode(zlib.compress(data))

def base64decode_and_gunzip(data):
  return zlib.decompress(base64.b64decode(data))

class SAMLValidationError(Exception):
  """SAML signature not validated"""
  def __init__(self, msg):
    self._msg = msg
  def __str__(self):
    return '%s: %s' % (self.__doc__, self._msg)

class SAMLConditionError(Exception):
  """SAML Condition not met"""
  def __init__(self, msg):
    self._msg = msg
  def __str__(self):
    return '%s: %s' % (self.__doc__, self._msg)

class SAMLNameIDError(Exception):
  """SAML NameID error"""
  def __init__(self, msg):
    self._msg = msg
  def __str__(self):
    return '%s: %s' % (self.__doc__, self._msg)

class Pysamlsp(object):
  def __init__(self, config = {}):
    self.ID = uuid.uuid4().hex
    self.assertion_consumer_service_url = config.get(
      'assertion_consumer_service_url'
    ) or ''
    self.issuer = config.get('issuer') or ''
    self.private_key = config.get('private_key') or ''
    self.sign_authnrequests = config.get('sign_authnrequests') or False
    self.certificate = config.get('certificate') or ''
  def samlp_maker(self):
    return ElementMaker(
      namespace='urn:oasis:names:tc:SAML:2.0:protocol',
      nsmap=dict(samlp='urn:oasis:names:tc:SAML:2.0:protocol'),
    )
  def saml_maker(self):
    return ElementMaker(
      namespace='urn:oasis:names:tc:SAML:2.0:assertion',
      nsmap=dict(saml='urn:oasis:names:tc:SAML:2.0:assertion'),
    )
  def authnrequest(self):
    authn_request = self.samlp_maker().AuthnRequest(
      ProtocolBinding = 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST',
      Version = '2.0',
      IssueInstant = iso_no_microseconds(datetime.utcnow()),
      ID = self.ID,
      AssertionConsumerServiceURL = self.assertion_consumer_service_url
    )
    authn_request.append(self.saml_maker().Issuer(self.issuer))
    authn_request.append(self.samlp_maker().NameIDPolicy(
      allowcreate='true',
      format='urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress'
    ))
    requested_authn_context = self.samlp_maker().RequestedAuthnContext(
      Comparison = "exact"
    )
    requested_authn_context_class_ref = \
      self.saml_maker().AuthnContextClassRef('urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport')
    requested_authn_context.append(requested_authn_context_class_ref)
    authn_request.append(requested_authn_context)
    return authn_request
  def authnrequest_as_string(self):
    return etree.tostring(self.authnrequest())
  def authnrequest_to_sign(self):
    authnrequest_to_sign = self.authnrequest()
    authnrequest_to_sign.append(
      etree.fromstring(XML_SIGNATURE_FRAGMENT))
    return etree.tostring(authnrequest_to_sign)
  def authnrequest_signed(self):
    tempfile = '/tmp/' + self.ID
    with open(tempfile, 'w') as fh:
      fh.write(self.authnrequest_to_sign())
    signed = xmlsec1(
      '--sign',
      '--privkey-pem', self.private_key,
      tempfile)
    os.remove(tempfile)
    return signed.stdout
  def redirect_for_idp(self):
    if self.sign_authnrequests:
      authnrequest = self.authnrequest_signed()
    else:
      authnrequest = self.authnrequest_as_string()
    return "%s%s%s" % (
      self.assertion_consumer_service_url,
      '&' if self.assertion_consumer_service_url.find('?') > 0 else '?',
      urllib.urlencode(
        [('SAMLRequest', gzip_and_base64encode(authnrequest))]
      )
    )
  def check_not_before_date(self, when):
    return datetime.strptime(when, '%Y-%m-%dT%H:%M:%SZ') < datetime.utcnow()
  def check_not_on_or_after_date(self, when):
    return datetime.strptime(when, '%Y-%m-%dT%H:%M:%SZ') >= datetime.utcnow()
  def response_as_xml(self, saml_response):
    parser = etree.XMLParser(remove_blank_text = True)
    response = etree.parse(StringIO(base64.b64decode(saml_response)), parser)
    return etree.tostring(response, pretty_print = True)
  def verify_signature(self, saml_response):
    tempfile = '/tmp/' + self.ID
    with open(tempfile, 'w') as fh:
      fh.write(base64.b64decode(saml_response).strip())
    try:
      verified = xmlsec1(
        '--verify',
        '--pubkey-cert-pem', self.certificate,
        '--store-references',
        '--id-attr:ID', 'urn:oasis:names:tc:SAML:2.0:assertion:Assertion',
        tempfile)
    except:
      raise
    finally:
      pass
      #os.remove(tempfile)
    if not (verified.exit_code == 0 and
        verified.stderr.find('SignedInfo References (ok/all): 1/1') > 0):
      raise SAMLValidationError('xmlsec1 error: %s' % verified.stderr)
  def user_is_valid(self, saml_response):
    response = etree.fromstring(base64.b64decode(saml_response).strip())
    try:
      nameid = response.xpath(
        '//saml:NameID',
        namespaces = {
          'saml': 'urn:oasis:names:tc:SAML:2.0:assertion',
          'samlp': 'urn:oasis:names:tc:SAML:2.0:protocol'})[0]
    except:
      raise SAMLNameIDError('NameID not given')
    try:
      condition = response.xpath(
        '/samlp:Response/saml:Assertion/saml:Conditions',
        namespaces = {
          'saml': 'urn:oasis:names:tc:SAML:2.0:assertion',
          'samlp': 'urn:oasis:names:tc:SAML:2.0:protocol'})[0]
    except:
      raise SAMLConditionError('Conditions not given in response')
    if not self.check_not_before_date(
        condition.attrib.get('NotBefore', '2012-12-31T00:00:00Z')):
      raise SAMLConditionError('NotBefore condition not met')
    if not self.check_not_on_or_after_date(
        condition.attrib.get('NotOnOrAfter', None)):
      raise SAMLConditionError('NotOnOrAfter condition not met')
    self.verify_signature(saml_response)
    return nameid.text.strip()



import unittest
from expecter import expect
from datetime import datetime
from lxml import etree
import string
import base64
from pysamlsp import *

def setUpmodule():
  pass

class TestUtilityFunctions(unittest.TestCase):
  def test_iso_now_no_microseconds(self):
    dt = datetime(2013, 4, 17, 14, 15, 53, 32711)
    expect(iso_no_microseconds(dt)) == '2013-04-17T14:15:53'

class TestPysamlspAuthnRequestRoot(unittest.TestCase):
  def test_the_root_element_namespace(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar.tag) == '{urn:oasis:names:tc:SAML:2.0:protocol}AuthnRequest'
  def test_the_root_ProtocolBinding_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar.get('ProtocolBinding')) == 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST'
  def test_the_root_Version_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar.get('Version')) == '2.0'
  def test_the_root_IssueInstant_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar.get('IssueInstant')) == iso_no_microseconds(datetime.utcnow())
  def test_the_root_ID_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(all(c in string.hexdigits for c in ar.get('ID')))
    expect(len(ar.get('ID'))) == 32
  def test_the_root_AssertionConsumerServiceURL_attribute(self):
    sp = Pysamlsp({'assertion_consumer_service_url': 'http://localhost'})
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar.get('AssertionConsumerServiceURL')) == 'http://localhost'

class TestPysamlspIssuerElement(unittest.TestCase):
  def test_the_element_namespace(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[0].tag) == '{urn:oasis:names:tc:SAML:2.0:assertion}Issuer'
  def test_the_element_text(self):
    sp = Pysamlsp({'issuer': 'http://localhost/SAML'})
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[0].text) == 'http://localhost/SAML'

class TestPysamlspNameIDPolicyElement(unittest.TestCase):
  def test_the_element_namespace(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[1].tag) == \
      '{urn:oasis:names:tc:SAML:2.0:protocol}NameIDPolicy'
  def test_the_root_allowcreate_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[1].get('allowcreate')) == 'true'
  def test_the_root_format_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[1].get('format')) == \
      'urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress'

class TestPysamlspRequestedAuthnContextElement(unittest.TestCase):
  def test_the_element_namespace(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[2].tag) == \
      '{urn:oasis:names:tc:SAML:2.0:protocol}RequestedAuthnContext'
  def test_the_element_Comparison_attribute(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[2].get('Comparison')) == 'exact'
  def test_the_child_element_text(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_as_string())
    expect(ar[2][0].text) == \
      'urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport'

class TestAuthnRequestToSign(unittest.TestCase):
  def test_signature_element(self):
    sp = Pysamlsp()
    ar = etree.fromstring(sp.authnrequest_to_sign())
    expect(
      ar.xpath( "ds:Signature",
        namespaces = {'ds': 'http://www.w3.org/2000/09/xmldsig#'})[0].tag
    ) == '{http://www.w3.org/2000/09/xmldsig#}Signature'

class TestAuthnSignedRequest(unittest.TestCase):
  def test_signed_element(self):
    sp = Pysamlsp({'private_key': os.path.dirname(os.path.abspath(__file__)) + '/support/saml_key.pem'})
    ar = etree.fromstring(sp.authnrequest_signed())
    expect(len(ar.xpath( "//ds:SignatureValue",
        namespaces = {'ds': 'http://www.w3.org/2000/09/xmldsig#'}))) > 0

class TestGzipAndBase64(unittest.TestCase):
  def test_gzip_and_base64(self):
    expect(
      base64decode_and_gunzip(
        gzip_and_base64encode("<root>Test</root>")
      )
    ) == "<root>Test</root>"

class TestRedirectForIdP(unittest.TestCase):
  def test_redirect(self):
    sp = Pysamlsp({
      'assertion_consumer_service_url': 'http://localhost',
      'sign_authnrequests': True,
      'private_key': os.path.dirname(os.path.abspath(__file__)) + '/support/saml_key.pem'
    })
    expect(
      sp.redirect_for_idp().\
        startswith('http://localhost?SAMLRequest=')) == True

# This test response must be signed each time it is changed. To sign it,
# find the file under support and change it there. Then pass it to xmlsec:
#   xmlsec1 --sign --pubkey-pem saml.pub --privkey-pem saml_key.pem \
#     samlresponse.xml > samlresponse_signed.xml
# Copy the contents of samlresponse_signed.xml to the block below.
TEST_SAML_RESPONSE = base64.b64encode("""
<?xml version="1.0"?>
<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" ID="identifier_2" InResponseTo="identifier_1" Version="2.0" IssueInstant="2004-12-05T09:22:05" Destination="https://sp.example.com/SAML2/SSO/POST">
  <saml:Issuer>https://idp.example.org/SAML2</saml:Issuer>
  <samlp:Status>
    <samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
  </samlp:Status>
  <saml:Assertion xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" ID="identifier_3" Version="2.0" IssueInstant="2004-12-05T09:22:05">
    <saml:Issuer>https://idp.example.org/SAML2</saml:Issuer>
    <saml:Subject>
      <saml:NameID Format="urn:oasis:names:tc:SAML:2.0:nameid-format:emailAddress">
        test@tartansolutions.com
      </saml:NameID>
      <saml:SubjectConfirmation Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">
        <saml:SubjectConfirmationData InResponseTo="identifier_1" Recipient="https://sp.example.com/SAML2/SSO/POST" NotOnOrAfter="2020-01-01T00:00:00"/>
      </saml:SubjectConfirmation>
    </saml:Subject>
    <saml:Conditions NotBefore="2012-12-31T00:00:00Z" NotOnOrAfter="2020-01-01T00:00:00Z">
      <saml:AudienceRestriction>
        <saml:Audience>https://sp.example.com/SAML2</saml:Audience>
      </saml:AudienceRestriction>
    </saml:Conditions>
    <saml:AuthnStatement AuthnInstant="2014-01-01T00:00:00" SessionIndex="identifier_3">
      <saml:AuthnContext>
        <saml:AuthnContextClassRef>
          urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport
       </saml:AuthnContextClassRef>
      </saml:AuthnContext>
    </saml:AuthnStatement>
  </saml:Assertion>
  <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#">
    <ds:SignedInfo> 
      <ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/> 
      <ds:SignatureMethod Algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"/>
      <ds:Reference>
        <ds:Transforms> 
          <ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/> 
          <ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/> 
        </ds:Transforms> 
        <ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/> 
        <ds:DigestValue>u1HBARnk8EV8ZAxThfa2WCu4tkI=</ds:DigestValue> 
      </ds:Reference> 
    </ds:SignedInfo> 
    <ds:SignatureValue>ZBCJqzmgELzOcS6HlY102s5i/Qa53/siPOMJjTOH68L5X5CcCxwF9JYWHqd0eXwd
TJ4V3tOHjX6vp5KF3DvptxS3g4GUxS2+5t1qZzNczNSZKPyNs/spFBwETSgGZxY3
cNOHd7tRERQ0y+cgyGFrpoxLmG5CedjABvMFhrmeYw4=</ds:SignatureValue> 
    <ds:KeyInfo> 
      <ds:X509Data/> 
    </ds:KeyInfo> 
  </ds:Signature>
</samlp:Response>
""")

TEST_NONSENSE_RESPONSE = base64.b64encode('<nonsense />')

class TestSAMLResponse(unittest.TestCase):
  def test_signature_verifies(self):
    sp = Pysamlsp({'certificate': os.path.dirname(os.path.abspath(__file__)) + '/support/saml.crt'})
    expect(sp.verify_signature(TEST_SAML_RESPONSE)) == None
  def test_signature_doesnt_verify(self):
    sp = Pysamlsp({'certificate': os.path.dirname(os.path.abspath(__file__)) + '/support/saml.crt'})
    with expect.raises():
      sp.verify_signature(TEST_NONSENSE_RESPONSE)
  def test_not_before_date_check(self):
    sp = Pysamlsp()
    expect(sp.check_not_before_date('2012-12-31T00:00:00Z')) == True
    expect(sp.check_not_before_date('2022-12-31T00:00:00Z')) == False
  def test_not_on_or_after_date_check(self):
    sp = Pysamlsp()
    expect(sp.check_not_on_or_after_date('2020-12-31T00:00:00Z')) == True
    expect(sp.check_not_on_or_after_date('2012-12-31T00:00:00Z')) == False
  def test_user_is_valid(self):
    sp = Pysamlsp({'certificate': os.path.dirname(os.path.abspath(__file__)) + '/support/saml.crt'})
    expect(sp.user_is_valid(TEST_SAML_RESPONSE)) == 'test@tartansolutions.com'
  def test_user_is_invalid(self):
    sp = Pysamlsp({'certificate': os.path.dirname(os.path.abspath(__file__)) + '/support/saml.crt'})
    with expect.raises(SAMLNameIDError):
      sp.user_is_valid(TEST_NONSENSE_RESPONSE)



from setuptools import setup
try:
  from sh import xmlsec1
except:
  print "IMPORTANT: the xmlsec1 system package is required by this library. Please remember to also install it through whatever means are appropriate on your operating system."


setup(
  name = 'pysamlsp',
  version = '0.1.2',
  author = "Rob Martin @version2beta",
  author_email = "rob@version2beta.com",
  description = "A service provider implementation for SAML2.0.",
  long_description = open('README.rst').read(),
  url='http://pypi.python.org/pypi/pysamlsp',
  license = "LICENSE.txt",
  keywords = "SAML2 SSO single sign on service provider only, for use with a remote IdP.",
  packages = ['pysamlsp'],
  install_requires = [
    'lxml'
  ],
  test_requires = [
    'nose',
    'expecter',
    'dingus'
  ],
  #data_files = { 'pysamlsp': ['pysamlsp/support/*'] },
  include_package_data = True
)

